@@ -78,7 +78,7 @@ static void check_unique_names(sexp* x) {
     return ;
   }
 
-  sexp* names = r_vec_names(x);
+  sexp* names = r_names(x);
   if (names == r_null) {
     r_abort("`data` must be uniquely named but does not have names");
   }
@@ -143,7 +143,7 @@ static sexp* tilde_fn = NULL;
 static sexp* restore_mask_fn = NULL;
 
 static void on_exit_restore_lexical_env(sexp* mask, sexp* old, sexp* frame) {
-  sexp* fn = KEEP(r_duplicate(restore_mask_fn, true));
+  sexp* fn = KEEP(r_clone(restore_mask_fn));
 
   sexp* env = KEEP(r_new_environment(r_base_env, 2));
   r_env_poke(env, mask_sym, mask);
@@ -209,13 +209,20 @@ static sexp* mask_find(sexp* env, sexp* sym) {
     // to a simple environment whose ancestry shouldn't be looked up.
     top_env = env;
   }
-  KEEP(top_env); // Help rchk
+  int n_kept = 0;
+  KEEP_N(top_env, n_kept);
 
   sexp* cur = env;
   do {
     sexp* obj = r_env_find(cur, sym);
-    if (obj != r_unbound_sym && !r_is_function(obj)) {
+    if (TYPEOF(obj) == PROMSXP) {
+      KEEP(obj);
+      obj = r_eval(obj, r_empty_env);
       FREE(1);
+    }
+
+    if (obj != r_unbound_sym && !r_is_function(obj)) {
+      FREE(n_kept);
       return obj;
     }
 
@@ -226,7 +233,7 @@ static sexp* mask_find(sexp* env, sexp* sym) {
     }
   } while (cur != r_empty_env);
 
-  FREE(1);
+  FREE(n_kept);
   return r_unbound_sym;
 }
 sexp* rlang_data_pronoun_get(sexp* pronoun, sexp* sym) {
@@ -296,7 +303,7 @@ sexp* rlang_as_data_mask(sexp* data) {
   case r_type_list: {
     check_unique_names(data);
 
-    sexp* names = r_vec_names(data);
+    sexp* names = r_names(data);
     bottom = KEEP_N(r_new_environment(r_empty_env, 0), n_protect);
 
     if (names != r_null) {
@@ -441,14 +448,14 @@ sexp* rlang_data_mask_clean(sexp* mask) {
   }
 
   // At this level we only want to remove our own stuff
-  r_env_unbind_all(mask, data_mask_objects_names, false);
+  r_env_unbind_strings(mask, data_mask_objects_names);
 
   // Remove everything in the other levels
   sexp* env = bottom;
   sexp* parent = r_env_parent(top);
   while (env != parent) {
     sexp* nms = KEEP(r_env_names(env));
-    r_env_unbind_names(env, nms, false);
+    r_env_unbind_names(env, nms);
     FREE(1);
     env = r_env_parent(env);
   }
