@@ -7,6 +7,8 @@
 #define r_names_sym R_NamesSymbol
 #define r_class_sym R_ClassSymbol
 #define r_dots_sym R_DotsSymbol
+#define r_namespace_sym R_DoubleColonSymbol
+#define r_namespace3_sym R_TripleColonSymbol
 
 extern sexp* r_dot_environment_sym;
 extern sexp* r_function_sym;
@@ -18,6 +20,10 @@ extern sexp* r_x_sym;
 extern sexp* r_y_sym;
 extern sexp* r_z_sym;
 
+extern sexp* r_dot_x_sym;
+extern sexp* r_dot_y_sym;
+extern sexp* r_dot_fn_sym;
+
 
 sexp* r_new_symbol(sexp* x, int* err);
 
@@ -32,6 +38,11 @@ static inline const char* r_sym_get_c_string(sexp* sym) {
   return CHAR(PRINTNAME(sym));
 }
 
+static inline sexp* r_sym_as_character(sexp* x) {
+  return Rf_ScalarString(PRINTNAME(x));
+}
+
+
 bool r_is_symbol(sexp* sym, const char* string);
 bool r_is_symbol_any(sexp* x, const char** strings, int n);
 
